//
// FPGA PACMAN "NTSC" video encoder for Xilinx SpartanII
//
// Version : beta2 (alpha version of NTSC)
//
// Copyright(c) 2003 Tatsuyuki Satoh , All rights reserved
//
// Important !
//
// This software is freeware for non-commercial use. 
// This software is provided "AS IS", without warranty of any kind.
// You can use this under your own risk. 
//
// 2003. 2.17 created
//

module venc_ntsc(
  pclk,dclk,vclk,rgb8_in,hs_in,vs_in,
  r_out,r_tri,g_out,g_tri,b_out,b_tri,
  hs_out,hs_tri,vs_out,vs_tri
);
// input signals
input pclk;             // 6.144MHz  input pixel clock
input dclk;             // 12.288Mhz output pixel clock
input vclk;		// VIDEO pixel clock (14.318MHz)
input [7:0] rgb8_in;    // RGB   input
input hs_in;            // HSYNC input (16KHz)
input vs_in;            // VSYNC input (60Hz)

// output signals
output [4:0] r_out,g_out,b_out; // R,G,B output level
output [4:0] r_tri,g_tri,b_tri; // R,G,B tristate control
output hs_out;          // HSYNC output
output hs_tri;          // HSYNC tristate control
output vs_out;          // VSYNC output
output vs_tri;          // VSYNC tristate control

/////////////////////////////////////////////////////////////////////////////
// setup parameter
/////////////////////////////////////////////////////////////////////////////

parameter YC_SEPARATE = 0; // 0=composite vide / other=Y/C separate

/////////////////////////////////////////////////////////////////////////////
// resample H line FIFO to 16KHz to 15KHz conversion
/////////////////////////////////////////////////////////////////////////////
wire [7:0] rgb_rs;
wire [2:0] red_rs   = rgb_rs[7:5];
wire [2:0] green_rs = rgb_rs[4:2];
wire [1:0] blue_rs  = rgb_rs[1:0];
wire hs_rs; // feedback HSYNC out from encoder
resample hline(.sclk(pclk),.shs(hs_in) ,.srgb8(rgb8_in),
               .dclk(vclk),.dhs(hs_rs),.drgb8(rgb_rs) );

/////////////////////////////////////////////////////////////////////////////
// 8BPP RGB to 24BPP RGB conversion
/////////////////////////////////////////////////////////////////////////////

wire [7:0] R = ({8{red_rs[2]}}   & 8'h97) + ({8{red_rs[1]}}   & 8'h47) + ({8{red_rs[0]}}   & 8'h21);
wire [7:0] G = ({8{green_rs[2]}} & 8'h97) + ({8{green_rs[1]}} & 8'h47) + ({8{green_rs[0]}} & 8'h21);
wire [7:0] B = ({8{blue_rs[1]}}  & 8'h97) + ({8{blue_rs[0]}}  & 8'h47);

/////////////////////////////////////////////////////////////////////////////
// RGB to YUV color space converter (with output DAC range adjust)
/////////////////////////////////////////////////////////////////////////////

wire [7:0] CY,CU,CV;
rgb_yuv  rgb_yuv1 (.R(R),.G(G),.B(B),.Y(CY),.U(CU),.V(CV));
wire [7:0] vout , yout , cout;

/////////////////////////////////////////////////////////////////////////////
// NTSC encoder (with output DAC range adjust)
/////////////////////////////////////////////////////////////////////////////
ntsc_enc video_out (
	.T4SC_I(vclk),
	.RESET(1'b0),
	.Y(CY) , .V(CV) , .U(CU),   // input
	.HS_I(hs_in), .VS_I(vs_in), // ext-sync signal
//output
	.VPOS_O(),
	.HPOS_O(),
	.FIELD(),
	.HSYNC_O(hs_rs),
	.VSYNC_O(),
	.CSYNC_O(),
	.Y_O(yout),
	.C_O(cout),
	.V_O(vout),
//debug
	.SC_O(),
	.CB_O(),
	.WIN_O()
);

/////////////////////////////////////////////////////////////////////////////
// output
/////////////////////////////////////////////////////////////////////////////

// C signal
wire [6:0] c_t = YC_SEPARATE ? { 7'b0000000 } : { 7'b1111111 };
wire [6:0] c_o = YC_SEPARATE ? cout[7:1]      : { 7'b0000000 };
assign {r_tri,hs_tri,vs_tri} = c_t;
assign {r_out,hs_out,vs_out} = c_o;

// composite Video / Y
wire [7:0] vy_out = YC_SEPARATE ? yout : vout;

assign {g_tri,b_tri} = { 10'b0000000011 };
assign {g_out,b_out} = { vy_out,2'b00   };

endmodule
