//
// FPGA PACMAN "VGA 8BPP" video encoder for Xilinx SpartanII
//
// Version : beta2
//
// Copyright(c) 2002,2003 Tatsuyuki Satoh , All rights reserved
//
// Important !
//
// This software is freeware for non-commercial use. 
// This software is provided "AS IS", without warranty of any kind.
// You can use this under your own risk. 
//
// 2004. 9.16 added generic dpram code
// 2003. 2.17 rebuild from 'dblscan.v'
//

//------------------------------------- Xilinx BSRAM select
`define XILINX_BSRAM

module venc_vga8(
  pclk,dclk,vclk,rgb8_in,hs_in,vs_in,
  r_out,r_tri,g_out,g_tri,b_out,b_tri,
  hs_out,hs_tri,vs_out,vs_tri
);
// input signals
input pclk;             // 6.144MHz  input pixel clock
input dclk;             // 12.288Mhz output pixel clock
input vclk;             // VIDEO pixel clock (not used)
input [7:0] rgb8_in;    // RGB   input
input hs_in;            // HSYNC input (16KHz)
input vs_in;            // VSYNC input (60Hz)

// output signals
output [4:0] r_out,g_out,b_out; // R,G,B output level
output [4:0] r_tri,g_tri,b_tri; // R,G,B tristate control
output hs_out;          // HSYNC output
output hs_tri;          // HSYNC tristate control
output vs_out;          // VSYNC output
output vs_tri;          // VSYNC tristate control

//---------------------------------------------------------------------------
// setup parameter
//---------------------------------------------------------------------------

parameter H_COUNT = 384; // number of pixels in H-SCAN
parameter HS_POS  = 16;  // HSYNC position 
parameter HS_COUNT= 8;   // HSYNC width / pixel
parameter VS_COUNT= 8;   // VSYNC width / HSYNC_OUT

//---------------------------------------------------------------------------
// input timming
//---------------------------------------------------------------------------
reg [8:0] hpos_i;    // input capture postion
reg ihs;
wire ic_raise = ~ihs&hs_in; // raise hsync
always @(posedge pclk)
begin
  hpos_i <= ic_raise ? 0 : hpos_i + 1;
  ihs    <=  hs_in;
end

//---------------------------------------------------------------------------
//output timming
//---------------------------------------------------------------------------
reg [8:0] hpos_o;
reg ohs , hsync_o;
wire oh_raise = ~ohs&hs_in; // raise hsync 
wire oh_return = oh_raise | (hpos_o == H_COUNT-1);
always @(posedge dclk)
begin
  hpos_o <= oh_return ? 0 : hpos_o + 1;
  ohs    <= hs_in;

  // remake HSYNC
  if(hpos_o == HS_POS)
  begin
    hsync_o <= 1'b1;
  end
  if(hpos_o == (HS_POS+HS_COUNT) )
    hsync_o <= 1'b0;
end

//---------------------------------------------------------------------------
// RGB capture(portA) & output(portB)
//---------------------------------------------------------------------------
`ifdef XILINX_BSRAM
// xilinx BSRAM
wire [7:0] rgb8_out;  // RGB output

RAMB4_S8_S8 double_scan_ram (
	.CLKA(pclk),.ADDRA(hpos_i),.DOA(),.DIA(rgb8_in)     ,.WEA(1'b1),.ENA(1'b1),.RSTA(1'b0),
	.CLKB(dclk),.ADDRB(hpos_o),.DOB(rgb8_out),.DIB(8'h00),.WEB(1'b0),.ENB(1'b1),.RSTB(1'b0));
`else
// generic dpram
reg [7:0] double_scan_ram[511:0];
always @(posedge pclk) double_scan_ram[hpos_i] <= rgb8_in;
wire [7:0] rgb8_out = double_scan_ram[hpos_o];
`endif

//---------------------------------------------------------------------------
// vsync remake
//
// 1 HSYNC_IN delay & HSYNC pulse width = 4xHSYNC(in)
//---------------------------------------------------------------------------

reg [2:0] vs_cnt;
reg vsync_o;

always @(posedge hsync_o)
begin
  if(~vs_in)
  begin
    vs_cnt <= VS_COUNT-1;
    vsync_o <= 1;
  end else begin
    if(vs_cnt==0)
      vsync_o <= 0;
    vs_cnt <= vs_cnt - 1;
  end
end

//---------------------------------------------------------------------------
// output
//---------------------------------------------------------------------------

assign r_tri = { 5'b00011 };
assign r_out = { rgb8_out[7:5],2'b00 };  // 470ohm,1Kohm,2Kohm,open,open
assign g_tri = { 5'b00011 };
assign g_out = { rgb8_out[4:2],2'b00 };  // 470ohm,1Kohm,2Kohm,open,open
assign b_tri = { 5'b00111 };
assign b_out = { rgb8_out[1:0],3'b000 }; // 470ohm,1Kohm,open,open,open

// converted SYNC signal
assign vs_tri = 1'b0;
assign vs_out = ~vsync_o;
assign hs_tri = 1'b0;
assign hs_out = ~hsync_o;

endmodule
