/////////////////////////////////////////////////////////////////////////////
// FPGA PACMAN internal memory module
//
// Version : 0.92
//
// Copyright(c) 2002-2004 Tatsuyuki Satoh , All rights reserved
//
// Important !
//
// This software is freeware for non-commercial use. 
// This software is provided "AS IS", without warranty of any kind.
// You can use this under your own risk. 
//
// 2004. 9.16 Created
//
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// internal memory
/////////////////////////////////////////////////////////////////////////////

//
// Internal RAMs used 'BlockSelectRAM'
//
// 1.tilemap code RAM
// 2.tilemap color RAM
// 3.work & object RAM
// 4.object position RAM
// 5.object line RAM
// 6.sound waveform ROM
// 7.color lockup table ROM
// 8.palette ROM
//
// External ROM
//
// 9.CPU code ROM
// 10.object pattern ROM
//

module im_gen(
//
  CLK3X,
//
  CFG_A,CFG_DI,CFG_WE,CFG_EN,
  
// Sahred Address & Data(Out) Bus
  AB,DB_I,DB_O,
  TILE_CS,TILE_WE, COL_CS,COL_WE, WORK_CS,WORK_WE,
  PWE,PDI,PDO,
  LA,LDO,LDI,LCS,LWE,
//  SA,S0DI,S0DO,S1DI,S1DO,SWE,
  WA,WDO,
  CLUTA,CLUT_DO,
  PAL_A,PAL_SEL,PAL_DO
);

/////////////////////////////////////////////////////////////////////////////
// module I/O assign
/////////////////////////////////////////////////////////////////////////////

input CLK3X;

input [15:0] CFG_A;
input [7:0] CFG_DI;
input CFG_WE;
input CFG_EN;

input [11:0] AB;
input [7:0] DB_I;
output [7:0] DB_O;

// TileCode RAM
input TILE_CS,TILE_WE;
// TileColor RAM
input COL_CS,COL_WE;
// Work/ObjectAttribute RAM
input WORK_CS,WORK_WE;
// Object Position RAM
input PWE;
input [7:0] PDI;
output [7:0] PDO;
// Object Line RAM
input [7:0] LA;
input [3:0] LDI;
output [3:0] LDO;
input LCS,LWE;
// sound Register RAM
// input [3:0] SA;
// output [3:0] S0DO , S1DO;
// input  [3:0] S0DI , S1DI;
// input SWE;
// sound WaveForm ROM
input [7:0] WA;
output [3:0] WDO;
//ColorLookup ROM
input [7:0] CLUTA;
output [3:0] CLUT_DO;
input [3:0] PAL_A;
input [1:0] PAL_SEL;
output [7:0] PAL_DO;

/////////////////////////////////////////////////////////////////////////////
// configration controll
/////////////////////////////////////////////////////////////////////////////
wire we4c    = CFG_WE & (CFG_A[15:10]==6'b010011);
wire we_wfm  = we4c & (CFG_A[9:8]==2'b00);  // 4c00-4cff : waveform rom
wire we_clut = we4c & (CFG_A[9:8]==2'b01);  // 4d00-4dff : clut rom
wire we_pal  = we4c & (CFG_A[9:8]==4'b10);  // 4e00-4eff : palette ROM

//wire [7:0] cfg_d = CFG_A[7:0];
wire [7:0] cfg_d = CFG_DI;

/////////////////////////////////////////////////////////////////////////////
// Video RAM
/////////////////////////////////////////////////////////////////////////////

//
// 1.tilemap code RAM
//
// Area    : 4000-43ff
// Size    : 8bit x 1024word
// Address : AB[9:0]
// Dout    : DB_O (shared)
// Din     : DB_I
// CS      : TILE_CS
// WE      : TILE_WE
// 
reg [7:0] tile_ram[1023:0];
reg [7:0] tile_do;
always @(posedge CLK3X)
begin
  if(~TILE_CS) tile_do <= 8'h00;
  else begin
    if(TILE_WE) tile_ram[AB[9:0]] <= DB_I;
    else        tile_do <= tile_ram[AB[9:0]];
  end
end

//
// 2.tilemap color RAM
//
// parts   : 
// Area    : 4400-47ff
// Size    : 8bit x 1024word
// Address : AB[9:0]
// Dout    : DB_O (shared)
// Din     : DB_I
// CS      : COL_CS
// WE      : COL_WE
// 
reg [7:0] col_ram[1023:0];
reg [7:0] col_do;
always @(posedge CLK3X)
begin
  if(~COL_CS) col_do <= 8'h00;
  else begin
    if(COL_WE) col_ram[AB[9:0]] <= DB_I;
    else       col_do <= col_ram[AB[9:0]];
  end
end

//
// 3.work & object RAM
//
// parts   : 4R,4M
// Area    : 4c00-4fff
// Size    : 8bit x 1024word
// Address : AB[9:0]
// Dout    : DB_O (shared)
// Din     : DB_I
// CS      : WORK_CS
// WE      : WORK_WE
// 
reg [7:0] work_ram[1023:0];
reg [7:0] work_do;
always @(posedge CLK3X)
begin
  if(~WORK_CS) work_do <= 8'h00;
  else begin
    if(WORK_WE) work_ram[AB[9:0]] <= DB_I;
    else        work_do <= work_ram[AB[9:0]];
  end
end

//
// multiplex vram data
//
assign DB_O = tile_do | col_do | work_do;


//
// 4.object position RAM
//
// parts   : 3F,3E
// Area    : 5060-506f
// Size    : 8bit x 16word
// Address : AB[3:0]
// Dout    : PDO (separate)
// Din     : PDI (separate)
// CS      : always select
// WE      : PWE
// 
reg [7:0] pram[15:0];
reg [7:0] PDO;
always @(posedge CLK3X)
begin
  if(PWE)  pram[AB[3:0]] <= PDI;
  else     PDO <= pram[AB[3:0]];
end

//
// 5.object line RAM
//
// parts   : (1A,1B,1C,1D) or (2A,2B,2B,2D)
// Area    : none
// Size    : 4bit x 256word
// Address : LA
// Dout    : LDO (separate)
// Din     : LDI (separate)
// CS      : LCS
// WE      : LWE (6MHz)
//
// note : LDO is latched LWE fall,so read data should be setup before LWE fall.
//
reg [3:0] lram[255:0];
reg [3:0] LDO;
always @(posedge CLK3X)
begin
  if(~LCS)  LDO <= 4'h0;
  else begin
    if(LWE) lram[LA] <= LDI;
    else    LDO <= lram[LA];
  end
end

//
// 6.sound waveform ROM
//
// parts   : 1M
// Area    : none
// Size    : 4bit x 256word
// Address : WA
// Dout    : WDO
// CS      : always select
//
reg [3:0] wfm_rom[255:0];
reg [3:0] WDO;
always @(posedge CLK3X)
begin
  if(we_wfm) wfm_rom[CFG_A[7:0]] <= cfg_d[3:0];
  else       WDO <= wfm_rom[WA];
end

//
// 7.color lockup table ROM
//
reg [3:0] clut_rom[255:0];
reg [3:0] CLUT_DO;
always @(posedge CLK3X)
begin
  if(we_clut) clut_rom[CFG_A[7:0]] <= cfg_d[3:0];
  else        CLUT_DO <= clut_rom[CLUTA];
end

//
// 8.palette ROM (RGB / HDTV / NTSC)
//
reg [7:0] pal_rom[15:0];
always @(posedge CLK3X)
begin
  if(we_pal) pal_rom[CFG_A[3:0]] <= cfg_d;
end
assign PAL_DO = pal_rom[PAL_A];

endmodule
