/////////////////////////////////////////////////////////////////////////////
// FPGA PACMAN internal memory module for Xilinx SpartanII
//
// Version : 0.92
//
// Copyright(c) 2002,2003 Tatsuyuki Satoh , All rights reserved
//
// Important !
//
// This software is freeware for non-commercial use. 
// This software is provided "AS IS", without warranty of any kind.
// You can use this under your own risk. 
//
// 2004. 9.16 rebuild
// 2003. 2. 6 beta2 Created
//
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// internal memory
/////////////////////////////////////////////////////////////////////////////

//
// Internal RAMs used 'BlockSelectRAM'
//
// 1.tilemap code RAM
// 2.tilemap color RAM
// 3.work & object RAM
// 4.object position RAM
// 5.object line RAM
// 6.sound waveform ROM
// 7.color lockup table ROM
// 8.palette ROM
//
// External ROM
//
// 9.CPU code ROM
// 10.object pattern ROM
//

module im_xc2s(
//
  CLK3X,
//
  CFG_A,CFG_DI,CFG_WE,CFG_EN,
  
// Sahred Address & Data(Out) Bus
  AB,DB_I,DB_O,
  TILE_CS,TILE_WE, COL_CS,COL_WE, WORK_CS,WORK_WE,
  PWE,PDI,PDO,
  LA,LDO,LDI,LCS,LWE,
//  SA,S0DI,S0DO,S1DI,S1DO,SWE,
  WA,WDO,
  CLUTA,CLUT_DO,
  PAL_A,PAL_SEL,PAL_DO
);

/////////////////////////////////////////////////////////////////////////////
// module I/O assign
/////////////////////////////////////////////////////////////////////////////
input CLK3X;

// ROM write
input [15:0] CFG_A;
input [7:0] CFG_DI;
input CFG_WE;
input CFG_EN;

//
input [11:0] AB;
input [7:0] DB_I;
output [7:0] DB_O;

// TileCode RAM
input TILE_CS,TILE_WE;
// TileColor RAM
input COL_CS,COL_WE;
// Work/ObjectAttribute RAM
input WORK_CS,WORK_WE;
// Object Position RAM
input PWE;
input [7:0] PDI;
output [7:0] PDO;
// Object Line RAM
input [7:0] LA;
input [3:0] LDI;
output [3:0] LDO;
input LCS,LWE;
// sound Register RAM
// input [3:0] SA;
// output [3:0] S0DO , S1DO;
// input  [3:0] S0DI , S1DI;
// input SWE;
// sound WaveForm ROM
input [7:0] WA;
output [3:0] WDO;
//ColorLookup ROM
input [7:0] CLUTA;
output [3:0] CLUT_DO;

input [3:0] PAL_A;
input [1:0] PAL_SEL;
output [7:0] PAL_DO;

/////////////////////////////////////////////////////////////////////////////
// configration controll
/////////////////////////////////////////////////////////////////////////////
wire we4c    = CFG_WE & (CFG_A[15:10]==6'b010011);
wire we_wfm  = we4c & (CFG_A[9:8]==2'b00);    // 4c00-4cff : waveform rom
wire we_clut = we4c & (CFG_A[9:8]==2'b01);    // 4d00-4dff : clut rom
wire we_pal  = we4c & (CFG_A[9:8]==4'b10);    // 4e00-4eff : palette ROM

wire [7:0] cd = CFG_DI;

/////////////////////////////////////////////////////////////////////////////
// Video RAM
//
// Video RAM is assigned to eight 3072x1bit BSRAM.
// Other memories are assigned to the back 1/4 of that BSRAM.
//
/////////////////////////////////////////////////////////////////////////////

wire v_we = TILE_WE | COL_WE | WORK_WE;
wire v_cs = TILE_CS | COL_CS | WORK_CS;

//
// 4.object position RAM
//
// parts   : 3F,3E
// Area    : 5060-506f
// Size    : 8bit x 16word
// Address : AB[3:0]
// Dout    : PDO (separate)
// Din     : PDI (separate)
// CS      : always select
// WE      : PWE
// 
wire [8:0] pa9  =  { 5'b10000,AB[3:0] };

//
// 5.object line RAM
//
// parts   : (1A,1B,1C,1D) or (2A,2B,2B,2D)
// Area    : none
// Size    : 4bit x 256word
// Address : LA
// Dout    : LDO (separate)
// Din     : LDI (separate)
// CS      : LCS
// WE      : LWE (6MHz)
//
// note : LDO is latched LWE fall,so read data should be setup before LWE fall.
//
wire [9:0] la10 = { 2'b10, LA };

//
// 6.sound waveform ROM
//
// parts   : 1M
// Area    : none
// Size    : 4bit x 256word
// Address : WA
// Dout    : WDO
// CS      : always select
//
wire [9:0] wfa10 = {2'b10,( CFG_EN ? CFG_A[7:0] : WA) };

//
// 7.color lockup table ROM
//
wire [9:0] cla10 = {2'b10, (CFG_EN ? CFG_A[7:0] : CLUTA) };

//
// 8.palette ROM
//
wire [8:0] pala9 = {3'b100,(CFG_EN ? CFG_A[5:0] : {PAL_SEL,PAL_A} ) };

//
// 9.HDTV encode ROM
//

//
// 10.NTSC encode ROM
//

/////////////////////////////////////////////////////////////////////////////
// RAM block
/////////////////////////////////////////////////////////////////////////////

// PRAM
RAMB4_S1_S8 RAM_D0 (
.CLKA(CLK3X),.ADDRA(AB),.DOA(DB_O[0]),.DIA(DB_I[0]),.ENA(1'b1),.WEA(v_we),.RSTA(~v_cs),
.CLKB(CLK3X),.ADDRB(pa9),.DOB(PDO),.DIB(PDI),.ENB(1'b1),.WEB(PWE),.RSTB(1'b0) );

// LRAM
RAMB4_S1_S4 RAM_D1 (
.CLKA(CLK3X),.ADDRA(AB),.DOA(DB_O[1]),.DIA(DB_I[1]),.ENA(1'b1),.WEA(v_we),.RSTA(~v_cs),
.CLKB(CLK3X),.ADDRB(la10),.DOB(LDO),.DIB(LDI),.ENB(LCS),.WEB(LWE),.RSTB(~LCS) );

// waveform ROM
RAMB4_S1_S4 RAM_D2 (
.CLKA(CLK3X),.ADDRA(AB),.DOA(DB_O[2]),.DIA(DB_I[2]),.ENA(1'b1),.WEA(v_we),.RSTA(~v_cs),
.CLKB(CLK3X),.ADDRB(wfa10),.DOB(WDO),.DIB(cd[3:0]),.ENB(1'b1),.WEB(we_wfm),.RSTB(1'b0) );

// CLUT ROM
RAMB4_S1_S4 RAM_D3 (
.CLKA(CLK3X),.ADDRA(AB),.DOA(DB_O[3]),.DIA(DB_I[3]),.ENA(1'b1),.WEA(v_we),.RSTA(~v_cs), 
.CLKB(CLK3X),.ADDRB(cla10),.DOB(CLUT_DO),.DIB(cd[3:0]),.ENB(1'b1),.WEB(we_clut),.RSTB(1'b0) );

// palette ROM
RAMB4_S1_S8 RAM_D4 (
.CLKA(CLK3X),.ADDRA(AB),.DOA(DB_O[4]),.DIA(DB_I[4]),.ENA(1'b1),.WEA(v_we),.RSTA(~v_cs),
.CLKB(CLK3X),.ADDRB(pala9),.DOB(PAL_DO),.DIB(cd),.ENB(1'b1),.WEB(we_pal),.RSTB(1'b0) );

RAMB4_S1_S1 RAM_D5 (
.CLKA(CLK3X),.ADDRA(AB),.DOA(DB_O[5]),.DIA(DB_I[5]),.ENA(1'b1),.WEA(v_we),.RSTA(~v_cs),
.CLKB(CLK3X),.ADDRB(0),.DOB(),.DIB(0),.ENB(1'b1),.WEB(1'b0),.RSTB(1'b0) );

RAMB4_S1_S1 RAM_D6 (
.CLKA(CLK3X),.ADDRA(AB),.DOA(DB_O[6]),.DIA(DB_I[6]),.ENA(1'b1),.WEA(v_we),.RSTA(~v_cs),
.CLKB(CLK3X),.ADDRB(0),.DOB(),.DIB(0),.ENB(1'b1),.WEB(1'b0),.RSTB(1'b0) );

RAMB4_S1_S1 RAM_D7 (
.CLKA(CLK3X),.ADDRA(AB),.DOA(DB_O[7]),.DIA(DB_I[7]),.ENA(1'b1),.WEA(v_we),.RSTA(~v_cs),
.CLKB(CLK3X),.ADDRB(0),.DOB(),.DIB(0),.ENB(1'b1),.WEB(1'b0),.RSTB(1'b0) );

endmodule
